module my_data
  implicit none
  INTEGER, ALLOCATABLE:: IAP(:), JA(:)
  complex, ALLOCATABLE:: A(:), DA(:,:)
  Integer mode
  integer, allocatable:: start_row(:)
  Integer m, n, nsv, ncv, opcount
  Double precision s0,s1
  INTEGER LWORK
  REAL, ALLOCATABLE:: BK(:,:)
  COMPLEX, ALLOCATABLE:: WORK(:),VK(:,:),UK(:,:),VPLUS(:),VM(:,:),UM(:,:)
end module my_data

SUBROUTINE SMALLCOND(LS,BD,BE)
  use my_data
  IMPLICIT NONE

  INTEGER LS,J
  REAL HALF, ONE, ZERO
  PARAMETER ( HALF=0.5E+0, ONE=1.0E+0, ZERO=0.0E+0 )
  REAL BD(ncv),BE(ncv),EPS,TOL,TMP1
  REAL TEN
  PARAMETER (TEN = 10.0E0)
  REAL SLAMCH
  EXTERNAL SLAMCH

  EPS = SLAMCH( 'Precision' )
  TOL = TEN*EPS

  IF (LS .EQ. 0) THEN

     TMP1 = ABS(BD(1))
     DO J = 1, ncv-1
        IF (ABS(BE(J)) .LE. TOL*TMP1) THEN
           BE(J) = ZERO
        ENDIF
        TMP1 = ABS(BD(J+1))*(TMP1/(TMP1+ABS(BE(J))))
     ENDDO

  ELSE

     TMP1 = ABS(BD(ncv))
     DO J = ncv-1, 1, -1
        IF (ABS(BE(J)) .LE. TOL*TMP1) THEN
           BE(J) = ZERO
        ENDIF
        TMP1 = ABS(BD(J))*(TMP1/(TMP1+ABS(BE(J))))
     ENDDO

  ENDIF

  RETURN

END SUBROUTINE SMALLCOND

SUBROUTINE STRF(Q,P,CPBK,BE)
  use my_data
  IMPLICIT NONE

  INTEGER I,J
  REAL Q(ncv,nsv),P(ncv,nsv),CPBK(ncv,ncv),BE(1,ncv)
  REAL CS1,SN1,CS2,SN2

  DO I=1,nsv
     DO J=ncv,I+1,-1
        CALL SLARTG(P(J-1,I),P(J,I),CS1,SN1,P(J-1,I))
        CALL SROT(nsv-I,P(J-1,I+1),ncv,P(J,I+1),ncv,CS1,SN1)
        CALL SROT(ncv,CPBK(1,J-1),1,CPBK(1,J),1,CS1,SN1)
        CALL CSROT(N,VK(1,J-1),1,VK(1,J),1,CS1,SN1)

        CALL SLARTG(Q(J-1,I),Q(J,I),CS2,SN2,Q(J-1,I))
        CALL SROT(nsv-I,Q(J-1,I+1),ncv,Q(J,I+1),ncv,CS2,SN2)
        CALL SROT(ncv,CPBK(J-1,1),ncv,CPBK(J,1),ncv,CS2,SN2)
        CALL CSROT(M,UK(1,J-1),1,UK(1,J),1,CS2,SN2)
        CALL SROT(1,BE(1,J-1),1,BE(1,J),1,CS2,SN2)
     ENDDO
  ENDDO

END SUBROUTINE STRF

SUBROUTINE STRF2(Q,P,CPBK,BE)
  use my_data
  IMPLICIT NONE

  INTEGER I,J
  REAL Q(ncv,nsv),P(ncv,nsv),CPBK(ncv,ncv),BE(1,ncv)
  REAL CS1,SN1,CS2,SN2

  DO I=1,nsv
     DO J=ncv,I+1,-1
        CALL SLARTG(Q(J-1,I),Q(J,I),CS2,SN2,Q(J-1,I))
        CALL SROT(nsv-I,Q(J-1,I+1),ncv,Q(J,I+1),ncv,CS2,SN2)
        CALL SROT(ncv,CPBK(1,J-1),1,CPBK(1,J),1,CS2,SN2)
        CALL CSROT(M,UK(1,J-1),1,UK(1,J),1,CS2,SN2)
        CALL SROT(1,BE(1,J-1),1,BE(1,J),1,CS2,SN2)

        CALL SLARTG(P(J-1,I),P(J,I),CS1,SN1,P(J-1,I))
        CALL SROT(nsv-I,P(J-1,I+1),ncv,P(J,I+1),ncv,CS1,SN1)
        CALL SROT(ncv,CPBK(J-1,1),ncv,CPBK(J,1),ncv,CS1,SN1)
        CALL CSROT(N,VK(1,J-1),1,VK(1,J),1,CS1,SN1)
     ENDDO
  ENDDO

END SUBROUTINE STRF2

SUBROUTINE STRF3(P,CPBK,BE)
  use my_data
  IMPLICIT NONE

  INTEGER I,J
  REAL P(ncv,nsv),CPBK(ncv,ncv),BE(1,ncv)
  REAL CS1,SN1,CS2,SN2

  DO I=1,nsv
     DO J=ncv,I+1,-1
        CALL SLARTG(P(J-1,I),P(J,I),CS1,SN1,P(J-1,I))
        CALL SROT(nsv-I,P(J-1,I+1),ncv,P(J,I+1),ncv,CS1,SN1)
        CALL SROT(ncv,CPBK(1,J-1),1,CPBK(1,J),1,CS1,SN1)
        CALL CSROT(N,VK(1,J-1),1,VK(1,J),1,CS1,SN1)
     ENDDO
  ENDDO

  DO I=1,nsv
     DO J=ncv,I+1,-1
        CALL SLARTG(CPBK(J-1,I),CPBK(J,I),CS2,SN2,CPBK(J-1,I))
        CALL SROT(nsv-I,CPBK(J-1,I+1),ncv,CPBK(J,I+1),ncv,CS2,SN2)
        CALL CSROT(M,UK(1,J-1),1,UK(1,J),1,CS2,SN2)
        CALL SROT(1,BE(1,J-1),1,BE(1,J),1,CS2,SN2)
     ENDDO
  ENDDO

END SUBROUTINE STRF3

SUBROUTINE STRF4(Q,CPBK,BE)
  use my_data
  IMPLICIT NONE

  INTEGER I,J
  REAL Q(ncv,nsv),CPBK(ncv,ncv),BE(1,ncv)
  REAL CS1,SN1,CS2,SN2

  DO I=1,nsv
     DO J=ncv,I+1,-1
        CALL SLARTG(Q(J-1,I),Q(J,I),CS1,SN1,Q(J-1,I))
        CALL SROT(nsv-I,Q(J-1,I+1),ncv,Q(J,I+1),ncv,CS1,SN1)
        CALL SROT(ncv,CPBK(1,J-1),1,CPBK(1,J),1,CS1,SN1)
        CALL CSROT(M,UK(1,J-1),1,UK(1,J),1,CS1,SN1)
        CALL SROT(1,BE(1,J-1),1,BE(1,J),1,CS1,SN1)
     ENDDO
  ENDDO

  DO I=1,nsv
     DO J=ncv,I+1,-1
        CALL SLARTG(CPBK(J-1,I),CPBK(J,I),CS2,SN2,CPBK(J-1,I))
        CALL SROT(nsv-I,CPBK(J-1,I+1),ncv,CPBK(J,I+1),ncv,CS2,SN2)
        CALL CSROT(N,VK(1,J-1),1,VK(1,J),1,CS2,SN2)
     ENDDO
  ENDDO

END SUBROUTINE STRF4

SUBROUTINE RESGKL(J,LS,METHOD)
  use my_data
  use omp_lib
  IMPLICIT NONE

  integer LS

  INTEGER I,J,IINFO,METHOD,RANK
  REAL SONE,SZERO,STWO
  PARAMETER(SONE=1.0E+0,SZERO=0.0E+0,STWO=2.0E+0)
  COMPLEX CONE,CZERO,CTWO
  PARAMETER(CONE=(1.0E+0,0.0E+0),CZERO=(0.0E+0,0.0E+0),CTWO=(2.0E+0,0.0E+0))
  REAL CDUMMY(1,1),BETA
  REAL CPBK(ncv,ncv),CPBK2(ncv,ncv),BD(ncv),BE(ncv)
  COMPLEX BDC(ncv)
  REAL WORK2(ncv*8)
  REAL Q(ncv,ncv),P(ncv,ncv)
  COMPLEX QC(ncv,ncv),PC(ncv,ncv)
  REAL SCNRM2,SDOT

  VK(1:N,J+1) = VPLUS(1:N)
  DO WHILE(J < ncv)
     CALL av(VK(1:N,J+1),UK(1:M,J+1))
     CALL CGS2(UK(1:M,J+1),UK,M,J)
     BK(J+1,J+1) = SCNRM2(M,UK(1:M,J+1),1)
     UK(1:M,J+1) = UK(1:M,J+1) / BK(J+1,J+1)
     CALL atv(UK(1:M,J+1), VPLUS)
     CALL CGS2(VPLUS,VK,N,J+1)

     BETA = SCNRM2(N,VPLUS,1)
     IF(J < ncv - 1) THEN
        BK(J+1,J+2) = BETA
        VK(1:N,J+2) = VPLUS(1:N) / BETA
     ELSE
        VPLUS(1:N) = VPLUS(1:N) / BETA
     END IF
     J = J+1
  END DO
  J = nsv

  Q = SZERO
  DO I = 1,ncv
     Q(I,I) = SONE
  END DO
  P = SZERO
  DO I = 1,ncv
     P(I,I) = SONE
  END DO

  IF ( METHOD == 1 ) THEN
     ! with lapack QR

     CALL SGEBRDG_4_BISIDE(nsv+1,BK,ncv,Q,P)
     DO I = 1,ncv
        BD(I) = BK(I,I)
     END DO
     DO I = 1,ncv-1
        BE(I) = BK(I,I+1)
     END DO

     IF (LS == 0) THEN

        CALL SBDSQRU( 'U',ncv,ncv,ncv,0,BD,BE,P,ncv,Q,ncv,CDUMMY,ncv,WORK,IINFO )

        QC=CMPLX(Q)
        CALL CGEMM('N','N',M,nsv,ncv,CONE,UK,M,QC,ncv,CZERO,UM,M)
        CALL CLACPY('A',M,nsv,UM,M,UK,M)
        PC=CMPLX(P)
        CALL CGEMM('N','C',N,nsv,ncv,CONE,VK,N,PC,ncv,CZERO,VM,N)
        CALL CLACPY('A',N,nsv,VM,N,VK,N)

        BK = SZERO
        DO I =1 ,nsv
           BK(I,I) = BD(I)
        END DO

        CALL SAXPY(nsv,BETA,Q(ncv,1),ncv,BK(1,nsv+1),1)

     ELSE

        CALL SBDSQRU( 'U',ncv,ncv,ncv,0,BD,BE,P,ncv,Q,ncv,CDUMMY,ncv,WORK,IINFO )

        DO I = 1, ncv/2
           CALL SSWAP(ncv,Q(1,ncv+1-I),1,Q(1,I),1) !Q(1:ncv,I) = Q(1:ncv,ncv+1-I)
        END DO
        DO I = 1, ncv/2
           CALL SSWAP(ncv,P(ncv+1-I,1),ncv,P(I,1),ncv) !P(I,1:ncv) = P(ncv+1-I,1:ncv)
        END DO

        QC=CMPLX(Q)
        CALL CGEMM('N','N',M,nsv,ncv,CONE,UK,M,QC,ncv,CZERO,UM,M)
        CALL CLACPY('A',M,nsv,UM,M,UK,M)
        PC=CMPLX(P)
        CALL CGEMM('N','C',N,nsv,ncv,CONE,VK,N,PC,ncv,CZERO,VM,N)
        CALL CLACPY('A',N,nsv,VM,N,VK,N)

        BK = SZERO
        DO I =1 ,nsv
           BK(I,I) = BD(ncv-I+1)
        END DO

        CALL SAXPY(nsv,BETA,Q(ncv,1),ncv,BK(1,nsv+1),1)

     ENDIF
     
  ELSE IF ( METHOD == 2 ) THEN
     ! with lapack QR
     ! reflector

     CPBK = BK
     CALL SGEBRDG_4_BISIDE(nsv+1,BK,ncv,Q,P)
     DO I = 1,ncv
        BD(I) = BK(I,I)
     END DO
     DO I = 1,ncv-1
        BE(I) = BK(I,I+1)
     END DO

     IF (LS == 0) THEN

        CALL SBDSQRU( 'U',ncv,ncv,ncv,0,BD,BE,P,ncv,Q,ncv,CDUMMY,ncv,WORK,IINFO )

        P = TRANSPOSE(P)
        
        CALL SGEQR2P(ncv,nsv,P,ncv,BD,WORK,IINFO)
        PC = CMPLX(P)
        BDC = CMPLX(BD)
        CALL CUNM2R('R','N',N,ncv,nsv,PC,ncv,BDC,VK,N,WORK,IINFO )
        CALL SORM2R('R','N',ncv,ncv,nsv,P,ncv,BD,CPBK,ncv,WORK,IINFO )

        CALL SGEQR2P(ncv,nsv,Q,ncv,BD,WORK,IINFO)
        QC = CMPLX(Q)
        BDC = CMPLX(BD)
        CALL CUNM2R('R','N',M,ncv,nsv,QC,ncv,BDC,UK,M,WORK,IINFO )
        CALL SORM2R('L','T',ncv,nsv,nsv,Q,ncv,BD,CPBK,ncv,WORK,IINFO )

        BE(1:ncv) = SZERO
        BE(ncv) = BETA
        CALL SORM2R('R','N',1,ncv,nsv,Q,ncv,BD,BE,1,WORK,IINFO )

        BK = SZERO
        DO I = 1,nsv
           BK(I,I)=CPBK(I,I)
        END DO

        CALL SCOPY(nsv,BE,1,BK(1,nsv+1),1)

     ELSE

        CALL SBDSQRU( 'U',ncv,ncv,ncv,0,BD,BE,P,ncv,Q,ncv,CDUMMY,ncv,WORK,IINFO )

        P = TRANSPOSE(P)
        
        DO I = 1, ncv/2
           CALL SSWAP(ncv,Q(1,ncv+1-I),1,Q(1,I),1) !Q(1:ncv,I) = Q(1:ncv,ncv+1-I)
        END DO
        DO I = 1, ncv/2
           CALL SSWAP(ncv,P(1,ncv+1-I),1,P(1,I),1) !P(1:ncv,I) = P(1:ncv,ncv+1-I)
        END DO

        CPBK = TRANSPOSE(CPBK)
        CALL SGEQR2P(ncv,nsv,Q,ncv,BD,WORK,IINFO)
        QC = CMPLX(Q)
        BDC = CMPLX(BD)
        CALL CUNM2R('R','N',M,ncv,nsv,QC,ncv,BDC,UK,M,WORK,IINFO )
        CALL SORM2R('R','N',ncv,ncv,nsv,Q,ncv,BD,CPBK,ncv,WORK,IINFO )
        BE(1:ncv) = SZERO
        BE(ncv) = BETA
        CALL SORM2R('R','N',1,ncv,nsv,Q,ncv,BD,BE,1,WORK,IINFO )

        CALL SGEQR2P(ncv,nsv,P,ncv,BD,WORK,IINFO)
        PC = CMPLX(P)
        BDC = CMPLX(BD)
        CALL CUNM2R('R','N',N,ncv,nsv,PC,ncv,BDC,VK,N,WORK,IINFO )
        CALL SORM2R('L','T',ncv,nsv,nsv,P,ncv,BD,CPBK,ncv,WORK,IINFO )

        BK = SZERO
        DO I = 1,nsv
           BK(I,I)=CPBK(I,I)
        END DO

        CALL SCOPY(nsv,BE,1,BK(1,nsv+1),1)

     ENDIF

  ELSE IF ( METHOD == 3 ) THEN
     ! with lapack QR
     ! Givens

     CPBK = BK
     CALL SGEBRDG_4_BISIDE(nsv+1,BK,ncv,Q,P)
     DO I = 1,ncv
        BD(I) = BK(I,I)
     END DO
     DO I = 1,ncv-1
        BE(I) = BK(I,I+1)
     END DO

     IF (LS == 0) THEN

        CALL SBDSQRU( 'U',ncv,ncv,ncv,0,BD,BE,P,ncv,Q,ncv,CDUMMY,ncv,WORK,IINFO )

        P = TRANSPOSE(P)

        BE(1:ncv) = SZERO
        BE(ncv) = BETA
        CALL STRF(Q,P,CPBK,BE)

        BK = SZERO
        DO I = 1,nsv
           BK(I,I)=CPBK(I,I)
        END DO

        CALL SCOPY(nsv,BE,1,BK(1,nsv+1),1)

     ELSE

        CALL SBDSQRU( 'U',ncv,ncv,ncv,0,BD,BE,P,ncv,Q,ncv,CDUMMY,ncv,WORK,IINFO )

        P = TRANSPOSE(P)

        DO I = 1, ncv/2
           CALL SSWAP(ncv,Q(1,ncv+1-I),1,Q(1,I),1) !Q(1:ncv,I) = Q(1:ncv,ncv+1-I)
        END DO
        DO I = 1, ncv/2
           CALL SSWAP(ncv,P(1,ncv+1-I),1,P(1,I),1) !P(1:ncv,I) = P(1:ncv,ncv+1-I)
        END DO

        BE(1:ncv) = SZERO
        BE(ncv) = BETA
        CPBK = TRANSPOSE(CPBK)
        CALL STRF2(Q,P,CPBK,BE)

        BK = SZERO
        DO I = 1,nsv
           BK(I,I)=CPBK(I,I)
        END DO

        CALL SCOPY(nsv,BE,1,BK(1,nsv+1),1)

     ENDIF

  ELSE IF ( METHOD == 4 ) THEN
     ! with lapack QR
     ! Givens

     CPBK = BK
     CALL SGEBRDG_4_BISIDE(nsv+1,BK,ncv,Q,P)
     DO I = 1,ncv
        BD(I) = BK(I,I)
     END DO
     DO I = 1,ncv-1
        BE(I) = BK(I,I+1)
     END DO

     IF (LS == 0) THEN

        CALL SBDSQRU( 'U',ncv,ncv,ncv,0,BD,BE,P,ncv,Q,ncv,CDUMMY,ncv,WORK,IINFO )

        P = TRANSPOSE(P)

        BE(1:ncv) = SZERO
        BE(ncv) = BETA
        CALL STRF3(P,CPBK,BE)

        BK = SZERO
        DO I = 1,nsv
           BK(I,I)=CPBK(I,I)
        END DO

        CALL SCOPY(nsv,BE,1,BK(1,nsv+1),1)

     ELSE

        CALL SBDSQRU( 'U',ncv,ncv,ncv,0,BD,BE,P,ncv,Q,ncv,CDUMMY,ncv,WORK,IINFO )

        DO I = 1, ncv/2
           CALL SSWAP(ncv,Q(1,ncv+1-I),1,Q(1,I),1) !Q(1:ncv,I) = Q(1:ncv,ncv+1-I)
        END DO

        BE(1:ncv) = SZERO
        BE(ncv) = BETA
        CPBK = TRANSPOSE(CPBK)
        CALL STRF4(Q,CPBK,BE)

        BK = SZERO
        DO I = 1,nsv
           BK(I,I)=CPBK(I,I)
        END DO

        CALL SCOPY(nsv,BE,1,BK(1,nsv+1),1)

     ENDIF

  ELSE IF ( METHOD == 5 ) THEN
     ! with lapack QR

     CPBK = BK
     CALL SGEBRDG_4_BISIDE(nsv+1,BK,ncv,Q,P)
     DO I = 1,ncv
        BD(I) = BK(I,I)
     END DO
     DO I = 1,ncv-1
        BE(I) = BK(I,I+1)
     END DO

     IF (LS == 0) THEN

        CALL SBDSQRU( 'U',ncv,ncv,ncv,0,BD,BE,P,ncv,Q,ncv,CDUMMY,ncv,WORK,IINFO )
        
        P = TRANSPOSE(P)

        CALL MGS(Q,WORK2)
        QC=CMPLX(Q)
        CALL CGEMM('N','N',M,nsv,ncv,CONE,UK,M,QC,ncv,CZERO,UM,M)
        CALL CLACPY('A',M,nsv,UM,M,UK,M)
        CALL MGS(P,WORK2)
        PC=CMPLX(P)
        CALL CGEMM('N','N',N,nsv,ncv,CONE,VK,N,PC,ncv,CZERO,VM,N)
        CALL CLACPY('A',N,nsv,VM,N,VK,N)

        CALL SGEMM('N','N',ncv,nsv,ncv,SONE,CPBK,ncv,P,ncv,SZERO,CPBK2,ncv)
        BK = SZERO
        DO I = 1,nsv
           BK(I,I)=SDOT(ncv,CPBK2(1,I),1,Q(1,I),1)
        END DO

        CALL SAXPY(nsv,BETA,Q(ncv,1),ncv,BK(1,nsv+1),1)

     ELSE

        CALL SBDSQRU( 'U',ncv,ncv,ncv,0,BD,BE,P,ncv,Q,ncv,CDUMMY,ncv,WORK,IINFO )
        
        P = TRANSPOSE(P)

        DO I = 1, ncv/2
           CALL SSWAP(ncv,Q(1,ncv+1-I),1,Q(1,I),1) !Q(1:ncv,I) = Q(1:ncv,ncv+1-I)
        END DO
        DO I = 1, ncv/2
           CALL SSWAP(ncv,P(1,ncv+1-I),1,P(1,I),1) !P(1:ncv,I) = P(1:ncv,ncv+1-I)
        END DO

        CALL MGS(Q,WORK2)
        QC=CMPLX(Q)
        CALL CGEMM('N','N',M,nsv,ncv,CONE,UK,M,QC,ncv,CZERO,UM,M)
        CALL CLACPY('A',M,nsv,UM,M,UK,M)
        CALL MGS(P,WORK2)
        PC=CMPLX(P)
        CALL CGEMM('N','N',N,nsv,ncv,CONE,VK,N,PC,ncv,CZERO,VM,N)
        CALL CLACPY('A',N,nsv,VM,N,VK,N)

        CALL SGEMM('T','N',ncv,nsv,ncv,SONE,CPBK,ncv,Q,ncv,SZERO,CPBK2,ncv)
        BK = SZERO
        DO I = 1,nsv
           BK(I,I)=SDOT(ncv,CPBK2(1,I),1,P(1,I),1)
        END DO

        CALL SAXPY(nsv,BETA,Q(ncv,1),ncv,BK(1,nsv+1),1)

     ENDIF
     
  ELSE IF ( METHOD == 6 ) THEN
     ! with OQDS

     CALL SGEBRDG_4_BISIDE(nsv+1,BK,ncv,Q,P)
     DO I = 1,ncv
        BD(I) = BK(I,I)
     END DO
     DO I = 1,ncv-1
        BE(I) = BK(I,I+1)
     END DO
     P = TRANSPOSE(P)
     
     IF (LS == 0) THEN
        CALL SOQDS( 'L',ncv,BD,BE,LS,RANK,P,ncv,Q,ncv,WORK,WORK2,IINFO )

        QC=CMPLX(Q)
        CALL CGEMM('N','N',M,nsv,ncv,CONE,UK,M,QC,ncv,CZERO,UM,M)
        CALL CLACPY('A',M,nsv,UM,M,UK,M)
        PC=CMPLX(P)
        CALL CGEMM('N','N',N,nsv,ncv,CONE,VK,N,PC,ncv,CZERO,VM,N)
        CALL CLACPY('A',N,nsv,VM,N,VK,N)

        BK = SZERO
        DO I =1 ,nsv
           BK(I,I) = BD(I)
        END DO
 
        CALL SAXPY(nsv,BETA,Q(ncv,1),ncv,BK(1,nsv+1),1)

        IF (RANK .LT. nsv) THEN
           DO I =1 ,nsv
              BK(I,nsv+1) = SZERO
           END DO
        ENDIF
        
     ELSE
        
        CALL SOQDS( 'U',ncv,BD,BE,LS,RANK,Q,ncv,P,ncv,WORK,WORK2,IINFO )

        DO I = 1, ncv/2
           CALL SSWAP(ncv,Q(1,ncv+1-I),1,Q(1,I),1) !Q(1:ncv,I) = Q(1:ncv,ncv+1-I)
        END DO
        DO I = 1, ncv/2
           CALL SSWAP(ncv,P(1,ncv+1-I),1,P(1,I),1) !P(1:ncv,I) = P(1:ncv,ncv+1-I)
        END DO

        QC=CMPLX(Q)
        CALL CGEMM('N','N',M,nsv,ncv,CONE,UK,M,QC,ncv,CZERO,UM,M)
        CALL CLACPY('A',M,nsv,UM,M,UK,M)
        PC=CMPLX(P)
        CALL CGEMM('N','N',N,nsv,ncv,CONE,VK,N,PC,ncv,CZERO,VM,N)
        CALL CLACPY('A',N,nsv,VM,N,VK,N)

        BK = SZERO
        DO I =1 ,nsv
           BK(I,I) = BD(ncv-I+1)
        END DO
        
        CALL SAXPY(nsv,BETA,Q(ncv,1),ncv,BK(1,nsv+1),1)
        
        IF (RANK .LT. ncv) THEN
           DO I =1 ,nsv
              BK(I,nsv+1) = SZERO
           END DO
        ENDIF
        
     ENDIF
     
  ELSE IF ( METHOD == 7 ) THEN
     ! with OQDS
     ! reflector

     CPBK = BK
     CALL SGEBRDG_4_BISIDE(nsv+1,BK,ncv,Q,P)
     DO I = 1,ncv
        BD(I) = BK(I,I)
     END DO
     DO I = 1,ncv-1
        BE(I) = BK(I,I+1)
     END DO
     P = TRANSPOSE(P)

     IF (LS == 0) THEN
        
        CALL SOQDS( 'L',ncv,BD,BE,LS,RANK,P,ncv,Q,ncv,WORK,WORK2,IINFO )

        CALL SGEQR2P(ncv,nsv,P,ncv,BD,WORK,IINFO)
        PC = CMPLX(P)
        BDC = CMPLX(BD)
        CALL CUNM2R('R','N',N,ncv,nsv,PC,ncv,BDC,VK,N,WORK,IINFO )
        CALL SORM2R('R','N',ncv,ncv,nsv,P,ncv,BD,CPBK,ncv,WORK,IINFO )
     
        CALL SGEQR2P(ncv,nsv,Q,ncv,BD,WORK,IINFO)
        QC = CMPLX(Q)
        BDC = CMPLX(BD)
        CALL CUNM2R('R','N',M,ncv,nsv,QC,ncv,BDC,UK,M,WORK,IINFO )
        CALL SORM2R('L','T',ncv,nsv,nsv,Q,ncv,BD,CPBK,ncv,WORK,IINFO )
     
        BE(1:ncv) = SZERO
        BE(ncv) = BETA
        CALL SORM2R('R','N',1,ncv,nsv,Q,ncv,BD,BE,1,WORK,IINFO ) 
     
        BK = SZERO
        DO I = 1,nsv
           BK(I,I)=CPBK(I,I)
        END DO

        CALL SCOPY(nsv,BE,1,BK(1,nsv+1),1)

        IF (RANK .LT. ncv) THEN
           DO I =1 ,nsv
              BK(I,nsv+1) = SZERO
           END DO
        ENDIF

     ELSE
        
        CALL SOQDS( 'U',ncv,BD,BE,LS,RANK,Q,ncv,P,ncv,WORK,WORK2,IINFO )

        DO I = 1, ncv/2
           CALL SSWAP(ncv,Q(1,ncv+1-I),1,Q(1,I),1) !Q(1:ncv,I) = Q(1:ncv,ncv+1-I)
        END DO
        DO I = 1, ncv/2
           CALL SSWAP(ncv,P(1,ncv+1-I),1,P(1,I),1) !P(1:ncv,I) = P(1:ncv,ncv+1-I)
        END DO

        CPBK = TRANSPOSE(CPBK)
        CALL SGEQR2P(ncv,nsv,Q,ncv,BD,WORK,IINFO)
        QC = CMPLX(Q)
        BDC = CMPLX(BD)
        CALL CUNM2R('R','N',M,ncv,nsv,QC,ncv,BDC,UK,M,WORK,IINFO )
        CALL SORM2R('R','N',ncv,ncv,nsv,Q,ncv,BD,CPBK,ncv,WORK,IINFO )
        BE(1:ncv) = SZERO
        BE(ncv) = BETA
        CALL SORM2R('R','N',1,ncv,nsv,Q,ncv,BD,BE,1,WORK,IINFO ) 
        
        CALL SGEQR2P(ncv,nsv,P,ncv,BD,WORK,IINFO)
        PC = CMPLX(P)
        BDC = CMPLX(BD)
        CALL CUNM2R('R','N',N,ncv,nsv,PC,ncv,BDC,VK,N,WORK,IINFO )
        CALL SORM2R('L','T',ncv,nsv,nsv,P,ncv,BD,CPBK,ncv,WORK,IINFO )
        
        BK = SZERO
        DO I = 1,nsv
           BK(I,I)=CPBK(I,I)
        END DO
     
        CALL SCOPY(nsv,BE,1,BK(1,nsv+1),1)

        IF (RANK .LT. ncv) THEN
           DO I =1 ,nsv
              BK(I,nsv+1) = SZERO
           END DO
        ENDIF
        
     ENDIF
     
  ELSE IF ( METHOD == 8 ) THEN
     ! with OQDS
     ! Givens

     CPBK = BK
     CALL SGEBRDG_4_BISIDE(nsv+1,BK,ncv,Q,P)
     DO I = 1,ncv
        BD(I) = BK(I,I)
     END DO
     DO I = 1,ncv-1
        BE(I) = BK(I,I+1)
     END DO
     P = TRANSPOSE(P)
     
     IF (LS == 0) THEN
        
        CALL SOQDS( 'L',ncv,BD,BE,LS,RANK,P,ncv,Q,ncv,WORK,WORK2,IINFO )

        BE(1:ncv) = SZERO
        BE(ncv) = BETA
        CALL STRF(Q,P,CPBK,BE)
     
        BK = SZERO
        DO I = 1,nsv
           BK(I,I)=CPBK(I,I)
        END DO

        CALL SCOPY(nsv,BE,1,BK(1,nsv+1),1)

        IF (RANK .LT. nsv) THEN
           DO I =1 ,nsv
              BK(I,nsv+1) = SZERO
           END DO
        ENDIF
        
     ELSE
        
        CALL SOQDS( 'U',ncv,BD,BE,LS,RANK,Q,ncv,P,ncv,WORK,WORK2,IINFO )
        
        DO I = 1, ncv/2
           CALL SSWAP(ncv,Q(1,ncv+1-I),1,Q(1,I),1) !Q(1:ncv,I) = Q(1:ncv,ncv+1-I)
        END DO
        DO I = 1, ncv/2
           CALL SSWAP(ncv,P(1,ncv+1-I),1,P(1,I),1) !P(1:ncv,I) = P(1:ncv,ncv+1-I)
        END DO

        BE(1:ncv) = SZERO
        BE(ncv) = BETA
        CPBK = TRANSPOSE(CPBK)
        CALL STRF2(Q,P,CPBK,BE)
     
        BK = SZERO
        DO I = 1,nsv
           BK(I,I)=CPBK(I,I)
        END DO

        CALL SCOPY(nsv,BE,1,BK(1,nsv+1),1)
        
        IF (RANK .LT. ncv) THEN
           DO I =1 ,nsv
              BK(I,nsv+1) = SZERO
           END DO
        ENDIF
        
     ENDIF

  ELSE IF ( METHOD == 9 ) THEN
     ! with OQDS
     ! Givens

     CPBK = BK
     CALL SGEBRDG_4_BISIDE(nsv+1,BK,ncv,Q,P)
     DO I = 1,ncv
        BD(I) = BK(I,I)
     END DO
     DO I = 1,ncv-1
        BE(I) = BK(I,I+1)
     END DO
     P = TRANSPOSE(P)
     
     IF (LS == 0) THEN
        
        CALL SOQDS( 'L',ncv,BD,BE,LS,RANK,P,ncv,Q,ncv,WORK,WORK2,IINFO )
        
        BE(1:ncv) = SZERO
        BE(ncv) = BETA
        CPBK = TRANSPOSE(CPBK)
        CALL STRF4(Q,CPBK,BE)
     
        BK = SZERO
        DO I = 1,nsv
           BK(I,I)=CPBK(I,I)
        END DO

        CALL SCOPY(nsv,BE,1,BK(1,nsv+1),1)

        IF (RANK .LT. nsv) THEN
           DO I =1 ,nsv
              BK(I,nsv+1) = SZERO
           END DO
        ENDIF
        
     ELSE
        
        CALL SOQDS( 'U',ncv,BD,BE,LS,RANK,Q,ncv,P,ncv,WORK,WORK2,IINFO )
        
        DO I = 1, ncv/2
           CALL SSWAP(ncv,P(1,ncv+1-I),1,P(1,I),1) !P(1:ncv,I) = P(1:ncv,ncv+1-I)
        END DO

        BE(1:ncv) = SZERO
        BE(ncv) = BETA
        CALL STRF3(P,CPBK,BE)
     
        BK = SZERO
        DO I = 1,nsv
           BK(I,I)=CPBK(I,I)
        END DO

        CALL SCOPY(nsv,BE,1,BK(1,nsv+1),1)
        
        IF (RANK .LT. ncv) THEN
           DO I =1 ,nsv
              BK(I,nsv+1) = SZERO
           END DO
        ENDIF
     
     ENDIF

  ELSE IF ( METHOD == 10 ) THEN
     ! with OQDS

     CPBK = BK
     CALL SGEBRDG_4_BISIDE(nsv+1,BK,ncv,Q,P)
     DO I = 1,ncv
        BD(I) = BK(I,I)
     END DO
     DO I = 1,ncv-1
        BE(I) = BK(I,I+1)
     END DO
     P = TRANSPOSE(P)

     IF (LS == 0) THEN
        
        CALL SOQDS( 'L',ncv,BD,BE,LS,RANK,P,ncv,Q,ncv,WORK,WORK2,IINFO )

        CALL MGS(Q,WORK2)
        QC=CMPLX(Q)
        CALL CGEMM('N','N',M,nsv,ncv,CONE,UK,M,QC,ncv,CZERO,UM,M)
        CALL CLACPY('A',M,nsv,UM,M,UK,M)
        CALL MGS(P,WORK2)
        PC=CMPLX(P)
        CALL CGEMM('N','N',N,nsv,ncv,CONE,VK,N,PC,ncv,CZERO,VM,N)
        CALL CLACPY('A',N,nsv,VM,N,VK,N)
        
        CALL SGEMM('N','N',ncv,nsv,ncv,SONE,CPBK,ncv,P,ncv,SZERO,CPBK2,ncv)
        BK = SZERO
        DO I = 1,nsv
           BK(I,I)=SDOT(ncv,CPBK2(1,I),1,Q(1,I),1)
        END DO
        
        CALL SAXPY(nsv,BETA,Q(ncv,1),ncv,BK(1,nsv+1),1)
        
        IF (RANK .LT. nsv) THEN
           DO I =1 ,nsv
              BK(I,nsv+1) = SZERO
           END DO
        ENDIF

     ELSE
        
        CALL SOQDS( 'U',ncv,BD,BE,LS,RANK,Q,ncv,P,ncv,WORK,WORK2,IINFO )

        DO I = 1, ncv/2
           CALL SSWAP(ncv,Q(1,ncv+1-I),1,Q(1,I),1) !Q(1:ncv,I) = Q(1:ncv,ncv+1-I)
        END DO
        DO I = 1, ncv/2
           CALL SSWAP(ncv,P(1,ncv+1-I),1,P(1,I),1) !P(1:ncv,I) = P(1:ncv,ncv+1-I)
        END DO

        CALL MGS(Q,WORK2)
        QC=CMPLX(Q)
        CALL CGEMM('N','N',M,nsv,ncv,CONE,UK,M,QC,ncv,CZERO,UM,M)
        CALL CLACPY('A',M,nsv,UM,M,UK,M)
        CALL MGS(P,WORK2)
        PC=CMPLX(P)
        CALL CGEMM('N','N',N,nsv,ncv,CONE,VK,N,PC,ncv,CZERO,VM,N)
        CALL CLACPY('A',N,nsv,VM,N,VK,N)
        
        CALL SGEMM('T','N',ncv,nsv,ncv,SONE,CPBK,ncv,Q,ncv,SZERO,CPBK2,ncv)
        BK = SZERO
        DO I = 1,nsv
           BK(I,I)=SDOT(ncv,CPBK2(1,I),1,P(1,I),1)
        END DO
     
        CALL SAXPY(nsv,BETA,Q(ncv,1),ncv,BK(1,nsv+1),1)

        IF (RANK .LT. ncv) THEN
           DO I =1 ,nsv
              BK(I,nsv+1) = SZERO
           END DO
        ENDIF
        
     ENDIF
     
  END IF

  RETURN
END SUBROUTINE RESGKL
!
Subroutine av(xin, yout)
  Use my_data
  Use omp_lib
  Implicit None
  Integer i, j,thr_num
  complex, Dimension(n), Intent(in) :: xin
  complex, Dimension(m), Intent(out) :: yout
  complex  czero
  parameter (czero=(0.0e0,0.0e0))
  EXTERNAL           CDOTU
  complex  CDOTU

  s1 = omp_get_wtime()
  if (mode .eq. 0) then

!$OMP PARALLEL PRIVATE(i,thr_num,j)
     thr_num = omp_get_thread_num()
     do i=start_row(thr_num+1),start_row(thr_num+2)-1
        yout(i) = czero
        do j=IAP(i),IAP(i+1)-1
           yout(i) = yout(i)+(A(j))*xin(JA(j))
        enddo
     enddo
!$OMP END PARALLEL
     
  else

!$OMP PARALLEL DO
     do i=1,M
        yout(i) = cdotu(n,DA(1,i),1,xin,1)
     enddo
!$OMP END PARALLEL DO

  end if
  s0 = s0 + (omp_get_wtime() - s1)
  
  opcount = opcount + 1
  return
End Subroutine av
!
Subroutine atv(xin, yout)
  Use my_data
  Use omp_lib
  Implicit None
  Integer i, j,thr_num
  complex, Dimension(m), Intent(in) :: xin
  complex, Dimension(n), Intent(out) :: yout
  complex  czero
  parameter (czero=(0.0e0,0.0e0))
  EXTERNAL           CDOTC
  complex  CDOTC

  s1 = omp_get_wtime()
  if (mode .eq. 0) then

     yout=czero

!$OMP PARALLEL PRIVATE(i,thr_num,j) REDUCTION(+:yout)
     thr_num = omp_get_thread_num()
     do i=start_row(thr_num+1),start_row(thr_num+2)-1
        do j=IAP(i),IAP(i+1)-1
           yout(JA(j))=yout(JA(j))+conjg(A(j))*xin(i)
        enddo
     enddo
!$OMP END PARALLEL
     
  else

     yout=czero

!$OMP PARALLEL DO REDUCTION(+:yout)
     do i=1,M
        do j =1,n
           yout(j)=yout(j)+conjg(DA(j,i))*xin(i)
        enddo
     enddo
!$OMP END PARALLEL DO

  end if
  s0 = s0 + (omp_get_wtime() - s1)

  opcount = opcount + 1
  return
End Subroutine atv









